"""Extract subchunks from PDF file."""

import base64
import operator
from functools import reduce
from pathlib import Path
from typing import Any, Callable

import fitz
import pymupdf

from unichunking.types import ChunkPosition, SubChunk

MIN_PIXELS: int = 128


def _handle_line(
    line: Any,
    width: float,
    height: float,
    subchunk_idx: int,
    page_num: int,
    file_name: str,
) -> tuple[list[SubChunk], int]:
    line_chunks: list[SubChunk] = []
    for span in line["spans"]:
        text = str(span["text"].replace("�", " ").strip())
        font: Any = span["font"]
        bbox: Any = span["bbox"]
        if "bold" in font.lower():
            text = f"**{text}**"
        if text:
            x0, y0, x1, y1 = bbox
            position = ChunkPosition(
                x0=x0 / width,
                y0=y0 / height,
                x1=x1 / width,
                y1=y1 / height,
            )
            line_chunks.append(
                SubChunk(
                    subchunk_id=subchunk_idx,
                    content=text,
                    page=page_num,
                    position=position,
                    file_name=file_name,
                ),
            )
            subchunk_idx += 1

    return line_chunks, subchunk_idx


async def _retrieve_subchunks(
    path: Path,
    status_manager: Any,
    function: Callable[[str], str],
) -> tuple[list[list[list[list[SubChunk]]]], list[SubChunk]]:
    chunks: list[list[list[list[SubChunk]]]] = []
    images_chunks: list[SubChunk] = []
    idx = 0
    image_idx = 0

    with pymupdf.Document(path) as doc:
        num_pages: Any = doc.page_count  # type: ignore
        for page_num in range(num_pages):
            # text
            if page_num % int(num_pages / 17 + 1) == 0:
                page_progress = int((page_num + 1) / num_pages * 75)
                await status_manager.update_status(
                    progress=page_progress,
                    start=status_manager.start,
                    end=status_manager.end,
                )
            page_chunks: list[list[list[SubChunk]]] = []
            textpage: Any = doc.load_page(page_num).get_textpage()  # type: ignore
            page = textpage.extractDICT(sort=False)
            width, height = page["width"], page["height"]
            blocks = page["blocks"]

            for block in blocks:
                block_chunks: list[list[SubChunk]] = []
                lines = block["lines"]
                for line in lines:
                    line_chunks, idx = _handle_line(
                        line=line,
                        width=width,
                        height=height,
                        subchunk_idx=idx,
                        page_num=page_num,
                        file_name=path.name,
                    )
                    if line_chunks:
                        block_chunks.append(line_chunks)
                if block_chunks:
                    page_chunks.append(block_chunks)
            if page_chunks:
                chunks.append(page_chunks)

            zoom = 1.0  # Modifier le facteur de zoom pour une meilleure qualité
            mat = fitz.Matrix(zoom, zoom)

            # Convertir la page en image pixmap
            page = doc.load_page(page_num)  # type: ignore
            pix = page.get_pixmap(matrix=mat)  # type: ignore
            image_bytes = pix.tobytes()  # type: ignore

            base64_image = base64.b64encode(image_bytes).decode("utf-8")  # type: ignore
            image_description = function(base64_image)
            if image_description:
                image_chunk = SubChunk(
                    subchunk_id=7_700_000 + image_idx,
                    content=image_description,
                    page=page_num,
                    position=ChunkPosition(
                        x0=0,
                        y0=0,
                        x1=1,
                        y1=1,
                    ),
                    file_name=path.name,
                    content_type="image",
                )
                image_idx += 1
                images_chunks.append(image_chunk)

    return chunks, images_chunks


def _filter_subchunks(
    chunks: list[list[list[list[SubChunk]]]],
) -> list[SubChunk]:
    flattened_chunks: list[SubChunk] = []

    for page_chunks in chunks:
        for block_chunks in page_chunks:
            for line_chunks in block_chunks:
                if line_chunks:
                    filtered_line_chunks = reduce(operator.add, line_chunks)
                    flattened_chunks.append(filtered_line_chunks)

    return flattened_chunks


async def extract_subchunks_pdf(
    path: Path,
    status_manager: Any,
    function: Callable[[str], str],
) -> list[SubChunk]:
    """Filetype-specific function : extracts subchunks from a PDF file.

    Args:
        path: Path to the local file.
        status_manager: Optional, special object to manage task progress.
        function: Function to handle images.

    Returns:
        A list of SubChunk objects.
    """
    chunks, images_chunks = await _retrieve_subchunks(
        path=path,
        status_manager=status_manager,
        function=function,
    )

    if len(images_chunks) > 0:
        flattened_chunks = images_chunks
    else:
        flattened_chunks = _filter_subchunks(chunks)

    progress = 100
    await status_manager.update_status(
        progress=progress,
        start=status_manager.start,
        end=status_manager.end,
    )

    return flattened_chunks



"""Extract subchunks from TXT/CSV/MD file."""

from pathlib import Path

from unichunking.tools import md_interpreter
from unichunking.types import (
    ChunkPosition,
    MatrixTable,
    SubChunk,
)


def extract_subchunks_txt(
    path: Path,
    is_csv: bool = False,
) -> tuple[list[SubChunk], list[MatrixTable], list[str]]:
    """Filetype-specific function : extracts subchunks from a TXT/MD/CSV file.

    Args:
        path: Path to the local file.
        is_csv: Boolean marker indicating whether file must be read as a CSV table.

    Returns:
        A tuple containing three lists:
        - List of subchunks, containing actual text or markers pointing to table/chart/image objects.
        - List of table/chart objects, of class MatrixTable.
        - List of image objects.
    """
    filename = path.name

    subchunks: list[SubChunk] = []
    tables: list[MatrixTable] = []
    images: list[str] = []

    if is_csv:
        table: list[list[str]] = []
        with path.open() as file:
            for line in file:
                line_values: list[str] = []
                cur_value = ""
                in_str = False
                line.replace("\n", "")
                for i in range(len(line)):
                    if line[i] == '"':
                        in_str = not in_str
                    elif line[i] == "," and not in_str:
                        line_values.append(cur_value)
                        cur_value = ""
                    else:
                        cur_value += line[i]
                line_values.append(cur_value)
                table.append(line_values)
        subchunks.append(
            SubChunk(
                subchunk_id=len(subchunks),
                content="0",
                page=0,
                position=ChunkPosition(0, 0, 0, 0),
                file_name=filename,
                content_type="table",
            ),
        )
        tables.append(MatrixTable("", table))

    else:
        with path.open() as file:
            lines = file.readlines()
            partial_subchunks, tables, images = md_interpreter(lines)
            for subchunk in partial_subchunks:
                subchunk.subchunk_id = len(subchunks)
                subchunk.file_name = filename
                subchunks.append(subchunk)

    return subchunks, tables, images



"""Surveil the status of a file processing task."""

from typing import Any

from unichunking.utils import logger


class TaskStatus:
    """Data class for storing task status information."""

    def __init__(
        self: "TaskStatus",
        task: str,
        progress: int = 0,
    ) -> None:
        """Initializes the TaskStatus object."""
        self.task = task
        self.progress = progress

    async def update_status(
        self: "TaskStatus",
        progress: int,
        start: int = 0,
        end: int = 100,
    ) -> None:
        """Update the progress of a TaskStatus object with a logging statement."""
        self.progress = (
            start if end == start else int(start + (progress / 100) * (end - start))
        )
        logger.debug(f'Task "{self.task}" updated progress to {self.progress}.')


class StatusManager:
    """Data class to handle status management."""

    def __init__(
        self: "StatusManager",
        status: Any = None,
        update_status_function: Any = None,
        task: str = "",
        start: int = 0,
        end: int = 100,
    ) -> None:
        """Initializes the StatusManager object."""
        if status and update_status_function:
            self.status = status
            self.update_status = update_status_function
        else:
            self.status = TaskStatus(task)
            self.update_status = self.status.update_status
        self.start = start
        self.end = end



"""Page computing."""

import re
from pathlib import Path
from typing import Any

import pymupdf

from unichunking.settings import unisettings
from unichunking.tools import convert_file
from unichunking.types import Chunk, StatusManager
from unichunking.utils import logger


def _strip_text(txt: str) -> str:
    image_description_pattern = r"@_@_@ IMAGE \d+ @_@_@"
    return (
        re.sub(image_description_pattern, "", txt)
        .replace("** Table : ", "")
        .replace(" ; End Of Table **", "")
        .replace("** Chart : ", "")
        .replace(" ; End Of Chart **", "")
        .replace(" ", "")
        .replace("\n", "")
        .replace("\xa0", "")
        .replace("-", "")
        .lower()
    )


def _intersection_of_length_n(
    chunk: str,
    block: str,
    intersection_length: int,
) -> bool:
    for start in range(len(block) - intersection_length + 1):
        if block[start : start + intersection_length] in chunk:
            return True
    return False


def _longest_intersection(
    chunk: str,
    block: str,
    min_intersection_length: int,
) -> int:
    floor = min_intersection_length
    ceiling = min(len(block), len(chunk))
    intersection_length = min_intersection_length
    while floor < ceiling:
        intersection_length = (floor + ceiling + 1) // 2
        if _intersection_of_length_n(
            chunk=chunk,
            block=block,
            intersection_length=intersection_length,
        ):
            floor = intersection_length
        else:
            ceiling = intersection_length - 1
    return (
        floor
        if _intersection_of_length_n(
            chunk=chunk,
            block=block,
            intersection_length=floor,
        )
        else 0
    )


def _positions_intersect(
    position1: dict[str, float],
    position2: dict[str, float],
) -> bool:
    return (
        (
            position1["x0"] <= position2["x0"] <= position1["x1"]
            and position1["y0"] <= position2["y0"] <= position1["y1"]
        )
        or (
            position1["x0"] <= position2["x1"] <= position1["x1"]
            and position1["y0"] <= position2["y0"] <= position1["y1"]
        )
        or (
            position1["x0"] <= position2["x0"] <= position1["x1"]
            and position1["y0"] <= position2["y1"] <= position1["y1"]
        )
        or (
            position1["x0"] <= position2["x1"] <= position1["x1"]
            and position1["y0"] <= position2["y1"] <= position1["y1"]
        )
    )


def _edit_chunk_positions(
    chunk: Chunk,
) -> None:
    different_positions: list[dict[str, float]] = []
    for position in chunk.metadata["positions"]:
        position["x0"] = max(0, position["x0"])
        position["y0"] = max(0, position["y0"])
        position["x1"] = min(1, position["x1"])
        position["y1"] = min(1, position["y1"])
        new = True
        for different_position in different_positions:
            if _positions_intersect(
                position,
                different_position,
            ) or _positions_intersect(different_position, position):
                different_position["x0"] = min(
                    different_position["x0"],
                    position["x0"],
                )
                different_position["y0"] = min(
                    different_position["y0"],
                    position["y0"],
                )
                different_position["x1"] = max(
                    different_position["x1"],
                    position["x1"],
                )
                different_position["y1"] = max(
                    different_position["y1"],
                    position["y1"],
                )
                new = False
                break
        if new:
            different_positions.append(position)
    chunk.metadata["positions"] = different_positions


def edit_positions_on_page(
    chunks_pages: list[list[Chunk]],
    extension: str,
) -> list[list[Chunk]]:
    """Adjusts the positions of subelements in Chunk objects to avoid overlaps.

    For DOCX : no highlighting because sourcing is not precise enough.

    Args:
        chunks_pages: A list of lists where chunks_pages[i][j] is Chunk j on page i.
        extension: File extension.

    Returns:
        The same list of lists of Chunk objects, whoses positions where adjusted.
    """
    if extension == "docx":
        for page in chunks_pages:
            for chunk in page:
                chunk.metadata["positions"] = [{"x0": 0, "y0": 0, "x1": 0, "y1": 0}]

    else:
        for page in chunks_pages:
            for chunk in page:
                _edit_chunk_positions(chunk)

    return chunks_pages


def _find_possible_pages(
    cur_page: int,
    pages_reach: int,
    pages: list[list[str]],
) -> list[int]:
    possible_pages: list[int] = [cur_page]
    for _ in range(pages_reach):
        for page_idx in range(possible_pages[-1] + 1, len(pages)):
            if any(block for block in pages[page_idx]):
                possible_pages.append(page_idx)
                return possible_pages
    return possible_pages


async def compute_pages(
    path: Path,
    chunks: list[Chunk],
    num_pages: int,
    status_manager: Any = None,
    increment_reach: bool = True,
) -> tuple[list[Chunk], int]:
    """Corrects the pagination of DOCX chunks by comparing the text to the converted PDF.

    Args:
        path: Path to the local file.
        chunks: List of Chunk objects.
        num_pages: Number of pages originally found through Page Breaks.
        status_manager: Optional, special object to manage task progress.
        increment_reach: Boolean value (True by default) indicating whether or not to increase the page reach when accumulating chunks on the same page.

    Returns:
        A tuple containing an int and a list:
        - The actual number of pages in the document.
        - A list containing the Chunk objects, updated with the correct page numbers.
    """
    if not status_manager:
        status_manager = StatusManager(task="Page computing")

    remove_local_file = False
    if not path.with_suffix(".pdf").exists():
        logger.debug("Converting DOCX-like file to PDF format...")
        convert_file(path, "pdf")
        remove_local_file = True
        logger.debug("PDF conversion done.")

    min_intersection_length = unisettings.chunking.MIN_INTERSECTION_LENGTH

    with pymupdf.Document(path.with_suffix(".pdf")) as doc:
        if num_pages == doc.page_count:  # type: ignore
            return chunks, num_pages

        logger.debug(
            f"Pdf.page_count ({doc.page_count}) does not match num_pages ({num_pages}).",  # type: ignore
        )
        logger.debug(
            f"Computing page numbers with minimal intersection length {min_intersection_length}...",
        )

        pages = [
            [_strip_text(block[4]) for block in page.get_text("blocks")]  # type: ignore
            for page in doc.pages()  # type: ignore
        ]

    best_page = 0
    pages_reach = 1

    for chunk_idx in range(len(chunks)):
        if chunk_idx % int(len(chunks) / 30 + 1) == 0:
            page_progress = int((chunk_idx + 1) / len(chunks) * 100)
            await status_manager.update_status(
                progress=page_progress,
                start=status_manager.start,
                end=status_manager.end,
            )

        chunk = chunks[chunk_idx]
        stripped_chunk = _strip_text(chunk.content)

        cur_page = min(max(best_page, chunk.metadata["page"] - 1), len(pages) - 1)

        best_intersection = sum(
            _longest_intersection(
                chunk=stripped_chunk,
                block=block,
                min_intersection_length=min_intersection_length,
            )
            for block in pages[cur_page]
        )
        best_page = cur_page

        possible_pages = _find_possible_pages(
            cur_page=cur_page,
            pages_reach=pages_reach,
            pages=pages,
        )

        for page_idx in possible_pages[1:]:
            page_intersection = sum(
                _longest_intersection(
                    chunk=stripped_chunk,
                    block=block,
                    min_intersection_length=min_intersection_length,
                )
                for block in pages[page_idx]
            )
            if page_intersection > best_intersection:
                best_intersection = page_intersection
                best_page = page_idx

        chunk.metadata["page"] = best_page + 1
        if increment_reach:
            pages_reach += 1 if best_page == cur_page else 0

    if remove_local_file:
        path.with_suffix(".pdf").unlink()
        logger.debug("Deleted local PDF file.")
    return chunks, len(pages)



"""File processing settings."""

from pydantic_settings import BaseSettings


class FilesSettings(BaseSettings):
    """Accepted file types and their associated content types."""

    ACCEPTED_FILE_TYPES: list[str] = [
        "pdf",
        "docx",
        "xlsx",
        "pptx",
        "doc",
        "xls",
        "ppt",
        "odt",
        "ods",
        "odp",
        "txt",
        "csv",
        "md",
        "ipynb",
    ]

    CONTENT_TYPES: dict[str, str] = {
        "pdf": "application/pdf",
        "docx": "application/docx",
        "doc": "application/docx",
        "odt": "application/docx",
        "xlsx": "application/xlsx",
        "xls": "application/xlsx",
        "ods": "application/xlsx",
        "pptx": "application/pptx",
        "ppt": "application/pptx",
        "odp": "application/pptx",
        "txt": "text/plain",
        "csv": "text/csv",
        "md": "text/markdown",
        "ipynb": "application/x-ipynb+json",
    }



"""Chunking settings."""

from pydantic_settings import BaseSettings


class ChunkingSettings(BaseSettings):
    """Default values for specific chunking tasks."""

    DEFAULT_K_MIN: int = 220
    DEFAULT_K_MAX: int = 1000
    DEFAULT_MIN_MAX_GAP: int = 100
    DEFAULT_OVERLAP: int = 0
    DEFAULT_PAGES_OVERLAP: int = 100

    FAKE_PAGES_LENGTH: int = 1000
    MIN_INTERSECTION_LENGTH: int = 3



"""Extract subchunks from IPYNB file."""

import contextlib
from pathlib import Path
from typing import Any

import nbformat as nbf

from unichunking.tools import md_interpreter
from unichunking.types import (
    ChunkPosition,
    MatrixTable,
    SubChunk,
)


def extract_subchunks_ipynb(
    path: Path,
) -> tuple[list[SubChunk], list[MatrixTable], list[str]]:
    """Filetype-specific function : extracts subchunks from a IPYNB file.

    Keeps a typed-cells structure, with information on the programming language.
    (ex : "Cell 1 (Text) : ...", "Cell 2 (Code: python 3.11.7) : ...", "Output 1 of cell 2 : ...", "Output 2 of cell 2 : ...").

    Args:
        path: Path to the local file.

    Returns:
        A tuple containing three lists:
        - List of subchunks, containing actual text or markers pointing to table/chart/image objects.
        - List of table/chart objects, of class MatrixTable.
        - List of image objects.
    """
    filename = path.name

    subchunks: list[SubChunk] = []
    tables: list[MatrixTable] = []
    images: list[str] = []

    notebook: Any = nbf.read(str(path), nbf.NO_CONVERT)  # type: ignore
    language = ""
    version = ""
    with contextlib.suppress(Exception):
        language = notebook.metadata.language_info.name
    with contextlib.suppress(Exception):
        version = notebook.metadata.language_info.version
    cells = notebook.cells

    for i_cell in range(len(cells)):
        match cells[i_cell].cell_type:
            case "markdown":
                subchunks.append(
                    SubChunk(
                        subchunk_id=len(subchunks),
                        content=f"** Cell {i_cell + 1} (Text) : ",
                        page=0,
                        position=ChunkPosition(0, 0, 0, 0),
                        file_name=filename,
                    ),
                )
                lines = cells[i_cell].source.split("\n")
                partial_subchunks, partial_tables, partial_images = md_interpreter(
                    lines,
                )

                for subchunk in partial_subchunks:
                    subchunk.subchunk_id = len(subchunks)
                    subchunk.file_name = filename
                    subchunks.append(subchunk)

                tables.extend(partial_tables)
                images.extend(partial_images)

                subchunks.append(
                    SubChunk(
                        subchunk_id=len(subchunks),
                        content=f" ; End of cell {i_cell + 1} ** ",
                        page=0,
                        position=ChunkPosition(0, 0, 0, 0),
                        file_name=filename,
                    ),
                )

            case "code":
                subchunks.append(
                    SubChunk(
                        subchunk_id=len(subchunks),
                        content=f"** Cell {i_cell + 1} (Code : {language} {version}) : ",
                        page=0,
                        position=ChunkPosition(0, 0, 0, 0),
                        file_name=filename,
                    ),
                )
                for line in cells[i_cell].source.split("\n"):
                    subchunks.append(
                        SubChunk(
                            subchunk_id=len(subchunks),
                            content=line + " ",
                            page=0,
                            position=ChunkPosition(0, 0, 0, 0),
                            file_name=filename,
                        ),
                    )
                subchunks.append(
                    SubChunk(
                        subchunk_id=len(subchunks),
                        content=f" ; End of cell {i_cell + 1} ** ",
                        page=0,
                        position=ChunkPosition(0, 0, 0, 0),
                        file_name=filename,
                    ),
                )
                for i_output in range(len(cells[i_cell].outputs)):
                    output_cell = cells[i_cell].outputs[i_output]
                    if output_cell.output_type != "stream":
                        continue
                    subchunks.append(
                        SubChunk(
                            subchunk_id=len(subchunks),
                            content=f"** Output {i_output + 1} of cell {i_cell + 1} : ",
                            page=0,
                            position=ChunkPosition(0, 0, 0, 0),
                            file_name=filename,
                        ),
                    )
                    for line in output_cell.text.split("\n"):
                        subchunks.append(
                            SubChunk(
                                subchunk_id=len(subchunks),
                                content=line + " ",
                                page=0,
                                position=ChunkPosition(0, 0, 0, 0),
                                file_name=filename,
                            ),
                        )
                    subchunks.append(
                        SubChunk(
                            subchunk_id=len(subchunks),
                            content=f" ; End of output {i_output + 1} of cell {i_cell + 1} ** ",
                            page=0,
                            position=ChunkPosition(0, 0, 0, 0),
                            file_name=filename,
                        ),
                    )

            case _:
                pass

    return subchunks, tables, images



"""Various useful tools for multi-format file processing."""

from .chunk_cleaning import clean_chunk
from .file_converter import convert_file
from .markdown_interpreter import md_interpreter
from .table_processing import compact_matrix
from .xml_tools import extract_charts, extract_media_xlsx

__all__ = [
    "clean_chunk",
    "compact_matrix",
    "convert_file",
    "extract_charts",
    "extract_media_xlsx",
    "md_interpreter",
]



"""Primary types used during subchunks extraction."""

import re
from typing import Literal


class ChunkPosition:
    """Class for handling positioning information during subchunks/chunks extraction."""

    x0: float
    y0: float
    x1: float
    y1: float

    def __init__(
        self: "ChunkPosition",
        x0: float,
        y0: float,
        x1: float,
        y1: float,
    ) -> None:
        """Create a ChunkPosition object."""
        self.x0 = x0
        self.y0 = y0
        self.x1 = x1
        self.y1 = y1

    def __add__(
        self: "ChunkPosition",
        other: "ChunkPosition",
    ) -> "ChunkPosition":
        """Adds two items' positions."""
        return ChunkPosition(
            x0=min(self.x0, other.x0),
            y0=min(self.y0, other.y0),
            x1=max(self.x1, other.x1),
            y1=max(self.y1, other.y1),
        )

    def to_dict(self: "ChunkPosition") -> dict[str, float]:
        """Convert ChunkPosition to a dictionary."""
        return {
            "x0": self.x0,
            "y0": self.y0,
            "x1": self.x1,
            "y1": self.y1,
        }


class SubChunk:
    """Class for handling subchunks extraction when reading different file formats."""

    subchunk_id: int
    content: str
    page: int
    position: ChunkPosition
    file_name: str
    content_type: Literal["text", "table", "image"]

    def __init__(
        self: "SubChunk",
        subchunk_id: int,
        content: str,
        page: int,
        position: ChunkPosition,
        file_name: str,
        content_type: Literal["text", "table", "image"] = "text",
    ) -> None:
        """Create a SubChunk object."""
        self.subchunk_id = subchunk_id
        self.content = content
        self.page = page
        self.position = position
        self.file_name = file_name
        self.content_type = content_type

    def __add__(self: "SubChunk", other: "SubChunk") -> "SubChunk":
        """Adds two SubChunk objects together."""
        content = self.content + " " + other.content
        re.sub(r"\s{2,}", " ", content)
        return SubChunk(
            subchunk_id=self.subchunk_id,
            content=content,
            page=min(self.page, other.page),
            position=self.position + other.position,
            file_name=self.file_name,
            content_type=self.content_type,
        )

    def __str__(self: "SubChunk") -> str:
        """Str representation of the object."""
        return self.content

    def __repr__(self: "SubChunk") -> str:
        """Str representation of the object."""
        return self.content + " " + self.content_type



"""Subchunk extraction and formating for any file type."""

from pathlib import Path
from typing import Any, Callable

from llmax import tokens

from unichunking.settings import unisettings
from unichunking.subchunk_extraction import (
    extract_subchunks_docx,
    extract_subchunks_ipynb,
    extract_subchunks_pdf,
    extract_subchunks_pptx,
    extract_subchunks_txt,
    extract_subchunks_xlsx,
)
from unichunking.types import MatrixTable, StatusManager, SubChunk
from unichunking.utils import logger


def _compute_page_num(
    cur_page_num: int,
    cur_page_size: int,
    subchunk_page: int,
    fake_pages_length: int,
) -> int:
    if fake_pages_length:
        return cur_page_num + (cur_page_size > fake_pages_length)
    return subchunk_page


def _split_table(
    table: MatrixTable,
    size_threshold: int,
) -> tuple[MatrixTable, MatrixTable]:
    table_matrix = table.content
    table_title = table.title

    sub_tab_1, sub_tab_2 = [], []

    first_line = table_matrix[0]
    sub_tab_1 = table_matrix[:2]
    split_after_line = 2
    while (
        tokens.count(str(table_matrix[: split_after_line + 1])) < size_threshold
        and split_after_line < len(table_matrix) - 1
    ):
        sub_tab_1 = table_matrix[: split_after_line + 1]
        split_after_line += 1
    sub_tab_2 = [first_line] + table_matrix[split_after_line:]

    return MatrixTable(table_title, sub_tab_1), MatrixTable(table_title, sub_tab_2)


def _format_table(
    table: MatrixTable,
    chunk_size: int,
    k_max: int,
    min_format: int = 2,
) -> list[str]:
    """Splits a table to fit in the k_max format.

    Keeps the columns/lines information on all subtables to avoid information loss.

    Args:
        table: MatrixTable object to split.
        chunk_size: Size of the current chunk without the table, in tokens.
        k_max: Maximum chunk size, in tokens.
        min_format: Minimum number of lines to keep in a subtable.

    Returns:
        A list of strings representing the subtables created.
    """
    if tokens.count(str(table.read())) + chunk_size <= k_max:
        return [table.read()]

    end_of_cur_chunk, leftover = _split_table(table, size_threshold=k_max - chunk_size)
    tables_done = [end_of_cur_chunk.read()]
    tables_temp = [leftover]

    while len(tables_temp) > 0:
        cur_table = tables_temp.pop()
        cur_table_str = cur_table.read()

        if cur_table.abort():
            tables_done.append(cur_table.read_as_str())

        elif (
            max(len(cur_table.content), len(cur_table.content[0])) <= min_format or tokens.count(cur_table_str) <= k_max
        ):
            tables_done.append(cur_table.read())

        else:
            sub_tab_1, sub_tab_2 = _split_table(cur_table, size_threshold=k_max)
            tables_done.append(sub_tab_1.read())
            tables_temp.append(sub_tab_2)

    return tables_done


async def _format_subchunks(
    subchunks: list[SubChunk],
    tables: list[MatrixTable],
    images: list[str],
    k_max: int,
    fake_pages_length: int,
    status_manager: Any = None,
) -> list[SubChunk]:
    """Processes the items extracted by filetype-specific functions to return SubChunks.

    Replaces markers to table/chart/image objects by the corresponding objects.
    Imposes k_max format by splitting subchunks of exceeding length.
    Handles artifical pagination for file types without clear page definition.

    Args:
        subchunks: List of subchhunks extracted by filetype-specific function.
        tables: List of table/chart objects extracted by filetype-specific function.
        images: List of image objects extracted by filetype-specific function.
        k_max: Maximum chunk size, in tokens.
        fake_pages_length: Maximum page size, in tokens, for file types without clear page definition.
        status_manager: Optional, special object to manage task progress.

    Returns:
        A list of SubChunk objects.
    """
    if fake_pages_length:
        logger.debug("Adding artificial page numbers.")

    chunks: list[SubChunk] = []
    cur_page_size = 0
    cur_page_num = 0
    subchunks_progress = 0

    for i_subchunk in range(len(subchunks)):
        new_progress = (int((i_subchunk + 1) / len(subchunks) * 100 / 5)) * 5
        if new_progress > subchunks_progress:
            subchunks_progress = new_progress
            await status_manager.update_status(
                progress=subchunks_progress,
                start=status_manager.start,
                end=status_manager.end,
            )

        subchunk_content = subchunks[i_subchunk].content
        subchunk_content_type = subchunks[i_subchunk].content_type

        match subchunk_content_type:
            case "table":
                table = tables[int(subchunk_content)]
                formatted_tables = _format_table(
                    table=table,
                    chunk_size=0,
                    k_max=k_max,
                )
                subchunk_content = formatted_tables.pop()
                for formatted_table in formatted_tables:
                    new_page_num = _compute_page_num(
                        cur_page_num=cur_page_num,
                        cur_page_size=cur_page_size,
                        subchunk_page=subchunks[i_subchunk].page,
                        fake_pages_length=fake_pages_length,
                    )
                    chunks.append(
                        SubChunk(
                            subchunk_id=len(chunks),
                            content=formatted_table,
                            page=new_page_num,
                            position=subchunks[i_subchunk].position,
                            file_name=subchunks[i_subchunk].file_name,
                            content_type="table",
                        ),
                    )
                    if new_page_num > cur_page_num:
                        cur_page_size = 0
                    cur_page_size += tokens.count(formatted_table)
                    cur_page_num = new_page_num

            case "image":
                subchunk_content = images[int(subchunk_content)]

            case _:
                pass

        if subchunk_content.strip():
            new_page_num = _compute_page_num(
                cur_page_num=cur_page_num,
                cur_page_size=cur_page_size,
                subchunk_page=subchunks[i_subchunk].page,
                fake_pages_length=fake_pages_length,
            )
            chunks.append(
                SubChunk(
                    subchunk_id=len(chunks),
                    content=subchunk_content,
                    page=new_page_num,
                    position=subchunks[i_subchunk].position,
                    file_name=subchunks[i_subchunk].file_name,
                    content_type=subchunks[i_subchunk].content_type,
                ),
            )
            if new_page_num > cur_page_num:
                cur_page_size = 0
            cur_page_size += tokens.count(chunks[-1].content)
            cur_page_num = new_page_num

    return chunks


async def extract_subchunks(
    path: Path,
    function: Callable[[str], str] = (lambda _: ""),
    k_max: int = unisettings.chunking.DEFAULT_K_MAX,
    status_manager: Any = None,
) -> list[SubChunk]:
    """Global function that extracts chunks from any document.

    Args:
        path: Path to the local file.
        function: Function to describe images
        k_max: Maximum chunk size, in tokens.
        status_manager: Optional, special object to manage task progress.

    Returns:
        A list of SubChunk objects.
    """
    if not status_manager:
        status_manager = StatusManager(task="Subchunk extraction")

    new_path = path
    extension = path.suffix[1:].lower()

    if extension == "pdf":
        formatted_subchunks = await extract_subchunks_pdf(
            path,
            status_manager=status_manager,
            function=function,
        )

    else:
        subchunks, tables, images = [], [], []
        fake_pages_length = unisettings.chunking.FAKE_PAGES_LENGTH

        match extension:
            case "docx" | "doc" | "odt":
                subchunks, tables, images, new_path = extract_subchunks_docx(
                    path,
                    extension,
                    function=function,
                )
                extension = "docx"
                fake_pages_length = 0

            case "xlsx" | "xls" | "ods":
                subchunks, tables, images, new_path = extract_subchunks_xlsx(
                    path,
                    extension,
                )

            case "pptx" | "ppt" | "odp":
                subchunks, tables, images, new_path = extract_subchunks_pptx(
                    path,
                    extension,
                    function=function,
                )
                fake_pages_length = 0

            case "txt" | "md" | "csv":
                subchunks, tables, images = extract_subchunks_txt(
                    path,
                    is_csv=extension == "csv",
                )

            case "ipynb":
                subchunks, tables, images = extract_subchunks_ipynb(path)

            case _:
                pass

        if new_path != path:
            new_path.unlink()
            logger.debug(
                "Deleted local converted file.",
            )

        formatted_subchunks = await _format_subchunks(
            subchunks=subchunks,
            tables=tables,
            images=images,
            k_max=k_max,
            fake_pages_length=fake_pages_length,
            status_manager=status_manager,
        )

    subchunks_final: list[SubChunk] = []

    for subchunk in formatted_subchunks:
        ct = subchunk.content
        ct = (
            ct.replace("\uf0b7", " ")
            .replace("\xe0", "à")
            .replace("\xe7", "ç")
            .replace("\xe8", "è")
            .replace("\xf4", "ô")
            .replace("\xe9", "é")
        )
        ct = ct.replace("   ", " ")
        subchunks_final.append(
            SubChunk(
                subchunk_id=subchunk.subchunk_id,
                content=ct,
                page=subchunk.page,
                position=subchunk.position,
                file_name=subchunk.file_name,
                content_type=subchunk.content_type,
            ),
        )

    logger.debug(
        "Subchunks extraction done.",
    )

    return subchunks_final



"""Algorithmic correction/processing of table structures."""

from unichunking.types import MatrixTable
from unichunking.utils import logger

list_of_known_units = ["", "€", "$", "£", "%", "(", ")", "[", "]", "{", "}", "/", "."]


def _clean_cell(cell: str) -> str:
    return (
        cell.replace(" ", "").replace("\n", "").replace("\xa0", "").replace("None", "")
    )


def _check_if_splitted_cell(table: MatrixTable) -> bool:
    """Identifies if the input table contains a single value, in which case it is probably the result of splitting a merged cell during XLSX processing."""
    value = ""
    found_something = False
    table_matrix = table.content
    for i in range(len(table_matrix)):
        for j in range(len(table_matrix[i])):
            if table_matrix[i][j] != value:
                if found_something:
                    return False
                value = table_matrix[i][j]
                found_something = True
    table.content = [[value]]
    return True


def _square_format(table: MatrixTable) -> bool:
    """If the table is not rectangular, adds empty cells to avoid index errors."""
    matrix_table = table.content
    for i in range(len(matrix_table)):
        for j in range(len(matrix_table[i])):
            temp = _clean_cell(matrix_table[i][j])
            if temp in list_of_known_units:
                matrix_table[i][j] = temp
        while len(matrix_table[i]) < len(matrix_table[0]):
            matrix_table[i].append("")
    table.content = matrix_table
    return True


def _delete_empty_lines(table: MatrixTable) -> bool:
    matrix_table = table.content
    for i in range(len(matrix_table) - 1, -1, -1):
        if not table.abort():
            is_empty_line = True
            j = 0
            while j < len(matrix_table[i]) and is_empty_line:
                is_empty_line = is_empty_line and not matrix_table[i][j]
                j += 1
            if is_empty_line:
                matrix_table.pop(i)
        else:
            break
        table.content = matrix_table
    return True


def _delete_empty_columns(table: MatrixTable) -> bool:
    matrix_table = table.content
    for j in range(len(matrix_table[0]) - 1, -1, -1):
        if not table.abort():
            is_empty_column = True
            i = 0
            while i < len(matrix_table) and is_empty_column:
                is_empty_column = is_empty_column and not matrix_table[i][j]
                i += 1
            if is_empty_column:
                for i in range(len(matrix_table)):
                    matrix_table[i].pop(j)
        else:
            break
        table.content = matrix_table
    return True


def _extract_title(table: MatrixTable) -> bool:
    """A first line that only contains 1 non-empty value (potentially repeated) is the result of python-docx splitting the table title in cells.

    Can be spread on multiple lines, hence the iteration.
    """
    matrix_table = table.content
    title_done = False
    while not (title_done or table.abort()):
        changed_something = False
        title = ""
        for j in range(len(matrix_table[0])):
            if matrix_table[0][j] not in {"", title}:
                if title:
                    changed_something = False
                    title_done = True
                    break
                title = matrix_table[0][j]
                changed_something = True
        title_done = title_done or not changed_something
        if changed_something or not title:
            table.title += title
            matrix_table.pop(0)
            table.content = matrix_table
        matrix_table = table.content
    table.content = matrix_table
    return not table.abort()


def _merge_lines(table: MatrixTable, base_line: int, added_line: int) -> bool:
    if not table.abort():
        matrix_table = table.content
        for k in range(len(matrix_table[base_line])):
            if matrix_table[base_line][k] != matrix_table[added_line][k]:
                matrix_table[base_line][k] += " " + str(matrix_table[added_line][k])
        matrix_table.pop(added_line)
        table.content = matrix_table
    return True


def _merge_columns(table: MatrixTable, base_column: int, added_column: int) -> bool:
    if not table.abort():
        matrix_table = table.content
        for k in range(len(matrix_table)):
            if matrix_table[k][base_column] != matrix_table[k][added_column]:
                matrix_table[k][base_column] += " " + str(matrix_table[k][added_column])
            matrix_table[k].pop(added_column)
        table.content = matrix_table
    return True


def _detect_unit(table: MatrixTable, column_1: int, column_2: int) -> bool:
    """Because of how python-docx splits merged cells, two columns can have the same header, one of them containing the units associated to the other one."""
    if not table.abort():
        matrix_table = table.content
        rate_1, rate_2 = 0, 0
        for k in range(1, len(matrix_table)):
            if matrix_table[k][column_1] == matrix_table[k][column_2]:
                rate_1 += 1
                rate_2 += 1
                continue
            keep_going = False
            if matrix_table[k][column_1] in list_of_known_units:
                rate_1 += 1
                keep_going = True
            if matrix_table[k][column_2] in list_of_known_units:
                rate_2 += 1
                keep_going = True
            if not keep_going:
                break
        if rate_1 == len(matrix_table) - 1:
            return _merge_columns(
                table=table,
                base_column=column_2,
                added_column=column_1,
            )
        if rate_2 == len(matrix_table) - 1:
            return _merge_columns(
                table=table,
                base_column=column_1,
                added_column=column_2,
            )
    return False


def _correct_double(table: MatrixTable, column_1: int, column_2: int) -> bool:
    """A column header being repeated can be the result of a duplicated column.

    Handles the case where python-docx sometimes duplicates a column's content with another column's header.
    """
    if not table.abort():
        matrix_table = table.content
        similarities = [
            [0 for _ in range(len(matrix_table[0]))],
            [0 for _ in range(len(matrix_table[0]))],
        ]
        for k in range(1, len(matrix_table)):
            for m in range(len(matrix_table[k])):
                similarities[0][m] += matrix_table[k][m] == matrix_table[k][column_1]
                similarities[1][m] += matrix_table[k][m] == matrix_table[k][column_2]
        for k in range(len(similarities[0])):
            if k != column_1 and similarities[0][k] == len(matrix_table) - 1:
                for b in range(len(matrix_table)):
                    matrix_table[b].pop(column_1)
                table.content = matrix_table
                return True
            if k != column_2 and similarities[1][k] == len(matrix_table) - 1:
                for m in range(len(matrix_table)):
                    matrix_table[m].pop(column_2)
                table.content = matrix_table
                return True
    return abs(column_1 - column_2) == 1 and _merge_lines(
        table=table,
        base_line=0,
        added_line=1,
    )


def _unit_detection_loop(table: MatrixTable) -> MatrixTable:
    """First auxiliary function to make merge_first_line_doubles less complex."""
    units_done = False
    while not (table.abort() or units_done):
        matrix_table = table.content
        changed_something = False
        j = 1
        while not table.abort() and j < len(matrix_table[0]) - 1:
            increment = True
            if matrix_table[0][j] and matrix_table[0][j] == matrix_table[0][j + 1]:
                increment = not _detect_unit(table=table, column_1=j, column_2=j + 1)
                changed_something = (not increment) or changed_something
            matrix_table = table.content
            if increment:
                j += 1
        units_done = not changed_something
    return table


def _double_correction_loop(table: MatrixTable) -> tuple[MatrixTable, bool]:
    """Second auxiliary function to make merge_first_line_doubles less complex."""
    matrix_table = table.content
    changed_something = False
    j = 1
    while not table.abort() and j < len(matrix_table[0]):
        increment = True
        if matrix_table[0][j]:
            for k in range(j):
                if matrix_table[0][k] == matrix_table[0][j]:
                    try:
                        changed_something = _correct_double(
                            table=table,
                            column_1=j,
                            column_2=k,
                        )
                        increment = not changed_something
                        if changed_something:
                            j -= 1
                            break
                    except IndexError as e:
                        logger.debug(f"An error occured during the table reading {e}")
            matrix_table = table.content
        if increment:
            j += 1
    return table, changed_something


def _merge_first_line_doubles(table: MatrixTable) -> bool:
    """A column header being repeated can also be caused by the first line having been splitted on multiple lines.

    This function merges lines/columns according to the situation until no repetition remains on the first line.
    """
    table = _unit_detection_loop(table)
    first_line_done = False
    while not (table.abort() or first_line_done):
        table, changed_something = _double_correction_loop(table)
        first_line_done = not changed_something
    return True


def _merge_fake_lines(table: MatrixTable) -> bool:
    """Lines are sometimes splitted on multiple lines, which can cause sub-lines to not have a header.

    This function merges lines that do not have a header to the previous line.
    """
    if not table.abort():
        matrix_table = table.content
        i = 1
        while not table.abort() and i < len(matrix_table):
            increment = True
            if not matrix_table[i][0]:
                _merge_lines(table=table, base_line=i - 1, added_line=i)
                increment = False
            matrix_table = table.content
            if increment:
                i += 1
    return True


def _find_fake_columns(
    matrix_table: list[list[str]],
    i: int,
    open_symbol: bool,
    merge_left: bool,
    fake_column: bool,
) -> tuple[bool, bool, bool]:
    """Identifies if a column without a header must be merged to the left or to the right."""
    for k in range(1, len(matrix_table)):
        if matrix_table[k][i] not in list_of_known_units:
            if matrix_table[k][i] == matrix_table[k][i - 1]:
                merge_left = True
            elif i < len(matrix_table[0]) - 1 and (
                matrix_table[k][i] == matrix_table[k][i + 1]
            ):
                merge_left = False
            else:
                fake_column = False
                break
        else:
            open_symbol = open_symbol or (matrix_table[k][i] in {"(", "[", "{"})
            merge_left = (
                merge_left
                and (not open_symbol)
                and (
                    len(matrix_table[k][i - 1].replace(" ", "")) > 0
                    and matrix_table[k][i - 1].replace(" ", "")[-1]
                    not in list_of_known_units
                )
            )
            if not (merge_left or open_symbol):
                merge_left = merge_left or (
                    i < len(matrix_table[0]) - 1
                    and (
                        len(matrix_table[k][i + 1].replace(" ", "")) > 0
                        and matrix_table[k][i + 1].replace(" ", "")[-1]
                        in list_of_known_units
                    )
                )
    return open_symbol, merge_left, fake_column


def _merge_fake_columns(table: MatrixTable) -> bool:
    """Identifies columns that do not have a header or are units columns and merges them to the left/right according to the situation."""
    if not table.abort():
        matrix_table = table.content
        default_merge_left = True
        i = 1
        while not table.abort() and i < len(matrix_table[0]):
            increment = True
            if matrix_table[0][i] in list_of_known_units:
                fake_column = True
                default_merge_left = default_merge_left and i > 1
                merge_left = default_merge_left
                open_symbol = False

                open_symbol, merge_left, fake_column = _find_fake_columns(
                    matrix_table=matrix_table,
                    i=i,
                    open_symbol=open_symbol,
                    merge_left=merge_left,
                    fake_column=fake_column,
                )

                if fake_column:
                    increment = False
                    if merge_left or i == len(matrix_table[0]) - 1:
                        _merge_columns(table=table, base_column=i - 1, added_column=i)
                    elif open_symbol:
                        _merge_columns(table=table, base_column=i, added_column=i + 1)
                    else:
                        _merge_columns(table=table, base_column=i + 1, added_column=i)

            matrix_table = table.content
            if increment:
                i += 1
    return True


def compact_matrix(
    table: MatrixTable,
    do_check_if_splitted_cell: bool = True,
    do_square_format: bool = True,
    do_delete_empty_lines: bool = True,
    do_delete_empty_columns: bool = True,
    do_extract_title: bool = True,
    do_merge_first_line_doubles: bool = True,
    do_merge_fake_lines: bool = True,
    do_merge_fake_columns: bool = True,
) -> MatrixTable:
    """Global table processing functions, calling all the optional subtasks according to the situation.

    All subtasks used for correction of DOCX tables.
    Only a few used for other formats.
    Reasoning behind some functions explained in their respective docstrings.

    Args:
        table: MatrixTable object to process.
        do_check_if_splitted_cell: Boolean marker indicating whether or not to check if the table is a splitted XLSX cell.
        do_square_format: Boolean marker indicating whether or not to add cells to make sure the table is a rectangle (same number of cells for each line/column).
        do_delete_empty_lines: Boolean marker indicating whether or not to delete all empty lines.
        do_delete_empty_columns: Boolean marker indicating whether or not to delete all empty columns.
        do_extract_title: Boolean marker indicating whether or not to try to extract a title from the first line doubles in the table.
        do_merge_first_line_doubles: Boolean marker indicating whether or not to merge lines/columns until there are no doubles on the first line.
        do_merge_fake_lines: Boolean marker indicating whether or not to merge lines that were wrongfully separated.
        do_merge_fake_columns: Boolean marker indicating whether or not to merge columns that only contain units to adjacent columns.

    Returns:
        The corrected MatrixTable object.
    """
    if not table.abort() and do_check_if_splitted_cell:
        _check_if_splitted_cell(table)

    if not table.abort() and do_square_format:
        _square_format(table)

    if not table.abort() and do_delete_empty_lines:
        _delete_empty_lines(table)
    if not table.abort() and do_delete_empty_columns:
        _delete_empty_columns(table)

    if not table.abort() and do_extract_title:
        _extract_title(table)

    if not table.abort() and do_merge_first_line_doubles:
        _merge_first_line_doubles(table)

    if not table.abort() and do_merge_fake_lines:
        _merge_fake_lines(table)
    if not table.abort() and do_merge_fake_columns:
        _merge_fake_columns(table)

    return table



"""Types used during file processing."""

from .chunking import Chunk
from .matrix_table import MatrixTable
from .status import StatusManager, TaskStatus
from .subchunking import ChunkPosition, SubChunk

__all__ = [
    "Chunk",
    "ChunkPosition",
    "MatrixTable",
    "StatusManager",
    "SubChunk",
    "TaskStatus",
]



"""Auxiliary functions to clean the text in a chunk."""

import re

from unstructured.nlp.patterns import UNICODE_BULLETS_RE

from unichunking.settings import unisettings
from unichunking.types import Chunk


def _clean_extra_whitespace(text: str) -> str:
    """Cleans extra whitespace characters that appear between words.

    Example:
    -------
    ITEM 1.     BUSINESS -> ITEM 1. BUSINESS
    """
    cleaned_text = re.sub(r"[\xa0]", " ", text)
    cleaned_text = re.sub(r"([ ]{2,})", " ", cleaned_text)
    return cleaned_text.strip()


def _clean_bullets(text: str) -> str:
    """Cleans unicode bullets from a section of text.

    Example:
    -------
    ●  This is an excellent point! -> This is an excellent point!
    """
    search = UNICODE_BULLETS_RE.match(text)
    if search is None:
        return text

    cleaned_text = UNICODE_BULLETS_RE.sub("", text, 1)
    return cleaned_text.strip()


def _clean(
    text: str,
) -> str:
    """Cleans text."""
    cleaned_text = _clean_extra_whitespace(text)
    cleaned_text = _clean_bullets(cleaned_text)
    return cleaned_text.strip()


def _clean_ligatures(text: str) -> str:
    """Replaces ligatures with their most likely equivalent characters.

    Example:
    -------
    The beneﬁts -> The benefits
    High quality ﬁnancial -> High quality financial
    """
    cleaned_text: str = text
    for k, v in unisettings.text.LIGATURES.items():
        cleaned_text = cleaned_text.replace(k, v)

    return cleaned_text


def _replace_unicode_quotes(text: str) -> str:
    r"""Replaces unicode bullets in text with the expected character.

    Example:
    -------
    \x93What a lovely quote!\x94 -> “What a lovely quote!”
    """
    for k, v in unisettings.text.UNICODE_QUOTES.items():
        text = text.replace(k, v)
    return text


def _clean_non_utf8_chars(text: str) -> str:
    r"""Cleans characters other than UTF-8 from a string.

    Example:
    -------
    b'\xc2\x88This text contains non-ascii characters!\xc2\x88'
        -> This text contains non-ascii characters!
    """
    utf8_encoded = text.encode("utf-8", "ignore")
    return utf8_encoded.decode("utf-8")


def clean_chunk(chunk: Chunk) -> Chunk:
    """Remove unwanted characters/blanks for better LLM readability."""
    content = chunk.content

    # Invoking Unstructured
    content = _replace_unicode_quotes(content)

    # CRUSHES FRENCH ACCENTS! ↓ :
    content = _clean_non_utf8_chars(content)

    content = _clean(content)
    content = _clean_ligatures(content)

    # Doing stuff my own way
    content = content.replace("\uf0b7", " ")
    content = content.replace("\xe0", "à")
    content = content.replace("\xe7", "ç")
    content = content.replace("\xe8", "è")
    content = content.replace("\xe9", "é")
    content = content.replace("\xf4", "ô")

    chunk.content = content

    return chunk



"""Chhunk formating."""

import copy
from pathlib import Path
from typing import Any, Callable, Literal
from uuid import uuid4

from llmax import tokens

from unichunking.settings import unisettings
from unichunking.tools import clean_chunk
from unichunking.types import Chunk, StatusManager, SubChunk

from .pages import compute_pages, edit_positions_on_page
from .subchunking import extract_subchunks
from .utils import compute_bbox

Action = Literal["describe", "transcribe"]


def split_chunks_with_overlap(
    chunks: list[SubChunk],
    page_id: int,
    file_name: str,
    file_hash: str = "",
    k_min: int = unisettings.chunking.DEFAULT_K_MIN,
    k_max: int = unisettings.chunking.DEFAULT_K_MIN
    + unisettings.chunking.DEFAULT_MIN_MAX_GAP,
    overlap: int = unisettings.chunking.DEFAULT_OVERLAP,
) -> list[Chunk]:
    """Splits the text allowing some overlapping.

    - chunks: les différents chunks extraits du document
    - k_min: nombre de tokens minimal par chunk
    - k_max: nombre maximal de tokens par chunk
    - overlap: nombre de caractères pour l'overlap.
    """
    if not (0 < k_min <= k_max) or overlap < 0:
        message = "Please ensure that 0 < K_MIN <= K_MAX and that overlap >= 0"
        raise ValueError(message)

    chunks_final: list[Chunk] = []

    current_chunk = ""
    current_positions = []

    for it, chunk in enumerate(chunks):
        if tokens.count(current_chunk) + tokens.count(
            chunk.content,
        ) < k_max or it + 1 == len(chunks):
            current_chunk += f" {chunk.content}"
            current_positions.append(chunk.position.to_dict())
            if tokens.count(current_chunk) >= k_min or (it + 1 == len(chunks)):
                curr_chunk = Chunk(
                    chunk_num=len(chunks_final) + 1,
                    content=current_chunk,
                    metadata={
                        "page": page_id,
                        "file_name": file_name,
                        "positions": copy.deepcopy(current_positions),
                    },
                    embedding=[0] * 1536,
                    chunk_uid=uuid4(),
                    file_hash=file_hash,
                    num_tokens=tokens.count(current_chunk),
                )
                curr_chunk.metadata["bbox"] = compute_bbox([curr_chunk])
                curr_chunk = clean_chunk(curr_chunk)
                if curr_chunk.content:
                    chunks_final.append(curr_chunk)
                current_positions = [chunk.position.to_dict()]
                current_chunk = current_chunk[-overlap:] if overlap else ""

        else:
            curr_chunk = Chunk(
                chunk_num=len(chunks_final) + 1,
                content=current_chunk,
                metadata={
                    "page": page_id,
                    "file_name": file_name,
                    "positions": copy.deepcopy(current_positions),
                },
                embedding=[0] * 1536,
                chunk_uid=uuid4(),
                file_hash=file_hash,
                num_tokens=tokens.count(current_chunk),
            )
            curr_chunk.metadata["bbox"] = compute_bbox([curr_chunk])
            curr_chunk = clean_chunk(curr_chunk)
            if curr_chunk.content:
                chunks_final.append(curr_chunk)
            current_positions = [chunk.position.to_dict()]
            current_chunk = (
                current_chunk[-overlap:] + chunk.content if overlap else chunk.content
            )

    return chunks_final


def default_function(base64_image: str) -> str:  # noqa: ARG001
    """Default function."""
    return ""


async def build_chunked_pages(
    path: Path,
    file_hash: str = "",
    k_min: int = unisettings.chunking.DEFAULT_K_MIN,
    overlap: int = unisettings.chunking.DEFAULT_OVERLAP,
    status_manager: Any = None,
    function: Callable[[str], str] = default_function,
) -> tuple[list[list[Chunk]], int]:
    """Returns a list of Pages and Chunks objects, once extracted, merged when necessary, cleaned and embedded."""
    k_max = k_min + unisettings.chunking.DEFAULT_MIN_MAX_GAP

    extension = path.suffix[1:].lower()
    if unisettings.files.CONTENT_TYPES[extension] == "application/docx":
        extension = "docx"

    chunks: list[Chunk] = []
    chunks_pages: list[list[Chunk]] = []
    file_name = path.name

    if not status_manager:
        status_manager = StatusManager(task="Subchunk extraction")
    status_manager.start = 0
    status_manager.end = 40

    chunks_extracted = await extract_subchunks(
        path,
        k_max=k_max,
        status_manager=status_manager,
        function=function,
    )

    pages_numbers = [ch.page for ch in chunks_extracted]
    if len(pages_numbers) == 0:
        return [[]], 0

    num_pages = max(pages_numbers) + 1
    status_manager.task = "Chunk formating"
    status_manager.start = 40
    status_manager.end = 75 if extension == "docx" else 100
    for page_id in range(num_pages):
        if page_id % int(num_pages / 30 + 1) == 0:
            page_progress = int((page_id + 1) / num_pages * 100)
            await status_manager.update_status(
                progress=page_progress,
                start=status_manager.start,
                end=status_manager.end,
            )

        relevant_chunks = [ch for ch in chunks_extracted if ch.page == page_id]
        res = split_chunks_with_overlap(
            relevant_chunks,
            file_hash=file_hash,
            file_name=file_name,
            page_id=page_id + 1,
            k_min=k_min,
            k_max=k_max,
            overlap=overlap,
        )
        chunks.extend(res)
        chunks_pages.append(res)

    if extension == "docx":
        status_manager.task = "Page computing"
        status_manager.start = 75
        status_manager.end = 100

        chunks, num_pages = await compute_pages(
            path=path,
            chunks=chunks,
            num_pages=num_pages,
            status_manager=status_manager,
        )
        chunks_pages: list[list[Chunk]] = [[] for _ in range(num_pages)]
        for chunk in chunks:
            chunks_pages[chunk.metadata["page"] - 1].append(chunk)  # type: ignore

    return edit_positions_on_page(chunks_pages, extension), num_pages
